import tp24.internal as internal
import tp24.model.colour as colour
from typing import Union

def gradient(a: colour.Colour, b: colour.Colour, ap: Union[int, float]=0.5, bp: Union[int, float]=0.5):
    b = internal.samemodel(a, b)
    av, bv = internal.tuplify(a, b)
    new = tuple(round((p*ap+q*bp)/(ap+bp)) for p, q in zip(av, bv))
    return internal.getclass(a, b)(*new)

def similarity(a: colour.Colour, b: colour.Colour):
    b = internal.samemodel(a, b)
    at = internal.unalpha(tuple(a), a)
    bt = internal.unalpha(tuple(b), b)
    ait = internal.unalpha(tuple(a.complementary()), a)

    if at == bt: return 1.0
    elif ait == bt: return 0.0

    scores = []
    for c in range(0, len(at)):
        minv = min(at[c], ait[c])
        maxv = max(at[c], ait[c])
        if (bt[c] == at[c]): scores.append(1)
        elif (bt[c] == ait[c]): scores.append(0)
        elif type(a).__name__.startswith("hs") and c == 0:
            #for degrees scales
            swap = False
            if bt[c] < minv:
                minv, maxv = maxv-360, minv
                swap = True
            elif bt[c] > maxv:
                minv, maxv = maxv, minv+360
                swap = True

            if (maxv == at[c] and not swap) or (maxv == at[c] and swap): #-ve gradient
                # m = (0-1)/(maxv-minv) = -1/(maxv-minv)
                # (p-1)/(bv-minv) = m
                # p-1 = m*(bv-minv)
                # p-1 = -1/(maxv-minv)*(bv-minv)
                # p = -1/(maxv-minv)*(bv-minv)+1
                scores.append(-1/(maxv-minv)*(bt[c]-minv)+1)
            elif (minv == at[c] and not swap) or (minv == at[c] and swap): #+ve gradient
                # m = (1-0)/(maxv-minv) = 1/(maxv-minv)
                # (p-0)/(bv-minv) = m
                # p = m*(bv-minv)
                # p = 1/(maxv-minv)*(bv-minv)
                scores.append(1/(maxv-minv)*(bt[c]-minv))
        else:
            if bt[c] < minv and minv == at[c]:
                # m = (1-.5)/minv = .5/minv
                # (p-.5)/(bv-0) = m
                # (p-.5)/bv = m
                # p-.5 = m*bv
                # p-.5 = .5/minv*bv
                # p = .5/min*bv + .5
                scores.append(0.5/minv*bt[c]+0.5)
            elif bt[c] < minv and minv == ait[c]:
                # m = (0-.5)/minv = -.5/minv
                # (p-.5)/(bv-0) = m
                # (p-.5)/bv = m
                # p-.5 = m*bv
                # p-.5 = -.5/minv*bv
                # p = -.5/min*bv + .5
                scores.append(-0.5/minv*bt[c]+0.5)
            elif bt[c] > maxv and maxv == at[c]:
                # m = (.5-1)/(maxr-maxv)) = -.5/(maxr-maxv))
                # (p-1)/(bv-maxv) = m
                # p-1 = m*(bv-maxv)
                # p-1 = -.5/(1-maxv)*(bv-maxv)
                # p = -.5/(maxr-maxv)*(bv-maxv)+1
                scores.append(-0.5/(a.RANGE[c]-maxv)*(bt[c]-maxv)+1)
            elif bt[c] > maxv and maxv == ait[c]:
                # m = (.5-0)/(maxr-maxv)) = .5/(maxr-maxv))
                # (p-0)/(bv-maxv) = m
                # p/(bv-maxv) = m
                # p = m*(bv-maxv)
                # p = .5/(maxr-maxv)*(bv-maxv)
                scores.append(0.5/(a.RANGE[c]-maxv)*(bt[c]-maxv))
            elif minv < bt[c] < maxv and minv == at[c]:
                # m = (0-1)/(maxv-minv) = -1/(maxv-minv)
                # (p-1)/(bv-minv) = m
                # p-1 = m*(bv-minv)
                # p-1 = -1/(maxv-minv)*(bv-minv)
                # p = -1/(maxv-minv)*(bv-minv)+1
                scores.append(-1/(maxv-minv)*(bt[c]-minv)+1)
            elif minv < bt[c] < maxv and minv == ait[c]:
                # m = (1-0)/(maxv-minv) = 1/(maxv-minv)
                # (p-0)/(bv-minv) = m
                # p = m*(bv-minv)
                # p = 1/(maxv-minv)*(bv-minv)
                scores.append(1/(maxv-minv)*(bt[c]-minv))
    
    return sum(scores)/len(scores)


import tp24.model.colour as colour
import tp24.model.m_rgb as col_rgb
import tp24.model.m_cmyk as col_cmyk
import tp24.errors as errors

class cmy(colour.Colour):
    RANGE = (255, 255, 255)
    c = None
    m = None
    y = None

    def __init__(self, vc: int, vm: int, vy: int):
        if not 0 <= vc <= self.RANGE[0]:
            raise errors.RangeError(f"Value of C channel is {vc} but is not in range of 0 <= c <= 255")
        elif not 0 <= vm <= self.RANGE[1]:
            raise errors.RangeError(f"Value of M channel is {vm} but is not in range of 0 <= m <= 255")
        elif not 0 <= vy <= self.RANGE[2]:
            raise errors.RangeError(f"Value of Y channel is {vy} but is not in range of 0 <= y <= 255")
        self.c = vc
        self.m = vm
        self.y = vy
    
    def __iter__(self):
        t = (self.c, self.m, self.y)
        for i in t:
            yield i

    def hsv(self):
        return self.rgb().hsv()

    def hsl(self):
        return self.rgb().hsl()

    def cmyk(self):
        c = self.c/255
        m = self.m/255
        y = self.y/255

        k = min(c, m, y)
        c *= k
        m *= k
        y -= k

        c *= 100
        m *= 100
        y *= 100

        if issubclass(type(self), colour.ColourAlpha):
            return col_cmyk.cmyka(c, m, y, k, self.a)
        else:
            return col_cmyk.cmyk(c, m, y, k)

    def rgb(self):
        r = 255 - self.c
        g = 255 - self.m
        b = 255 - self.y

        if issubclass(type(self), colour.ColourAlpha):
            return col_rgb.rgba(r, g, b, self.a)
        else:
            return col_rgb.rgb(r, g, b)

class cmya(cmy, colour.ColourAlpha):
    def __init__(self, *v):
        cmy.__init__(self, *v[:-1])
        colour.ColourAlpha.__init__(self, v[-1])

    def __iter__(self):
        t = (self.c, self.m, self.y, self.a)
        for i in t:
            yield i


import tp24.model.colour as colour
import tp24.model.m_cmyk as col_cmyk
import tp24.model.m_cmy as col_cmy
import tp24.model.m_hsv as col_hsv
import tp24.model.m_hsl as col_hsl
import tp24.errors as errors

class rgb(colour.Colour):
    RANGE = (255, 255, 255)
    r = None
    g = None
    b = None

    def __init__(self, vr: int, vg: int, vb: int):
        if not 0 <= vr <= self.RANGE[0]:
            raise errors.RangeError(f"Value of R channel is {vr} but is not in range of 0 <= r <= 255")
        elif not 0 <= vg <= self.RANGE[1]:
            raise errors.RangeError(f"Value of G channel is {vg} but is not in range of 0 <= g <= 255")
        elif not 0 <= vb <= self.RANGE[2]:
            raise errors.RangeError(f"Value of b channel is {vb} but is not in range of 0 <= b <= 255")
        self.r = vr
        self.g = vg
        self.b = vb
    
    def __iter__(self):
        t = (self.r, self.g, self.b)
        for i in t:
            yield i

    def hsv(self):
        r = self.r/self.RANGE[0]
        g = self.g/self.RANGE[1]
        b = self.b/self.RANGE[2]
        cmax = max(r, g, b)
        cmin = min(r, g, b)
        d = cmax-cmin

        if d == 0: h = 0
        elif cmax == r: h = 60 * (((g-b)/d)%6)
        elif cmax == g: h = 60 * (((b-r)/d)+2)
        elif cmax == b: h = 60 * (((r-g)/d)+4)

        if cmax == 0: s = 0
        else: s = d / cmax

        v = cmax

        h = round(h)
        s = round(s*100)
        v = round(v*100)
        
        if issubclass(type(self), colour.ColourAlpha):
            return col_hsv.hsva(h, s, v, self.a)
        else:
            return col_hsv.hsv(h, s, v)

    def hsl(self):
        r = self.r/self.RANGE[0]
        g = self.g/self.RANGE[1]
        b = self.b/self.RANGE[2]
        cmax = max(r, g, b)
        cmin = min(r, g, b)
        d = cmax-cmin

        if d == 0: h = 0
        elif cmax == r: h = 60 * (((g-b)/d)%6)
        elif cmax == g: h = 60 * (((b-r)/d)+2)
        elif cmax == b: h = 60 * (((r-g)/d)+4)
        
        l = (cmax+cmin)/2
        
        if d == 0: s = 0
        else: s = d / (1-abs(2*l-1))
        
        h = round(h)
        s = round(s*100)
        l = round(l*100)
        
        if issubclass(type(self), colour.ColourAlpha):
            return col_hsl.hsla(h, s, l, self.a)
        else:
            return col_hsl.hsl(h, s, l)


    def cmyk(self):
        r = self.r/self.RANGE[0]
        g = self.g/self.RANGE[1]
        b = self.b/self.RANGE[2]
        
        k = 1-max(r, g, b)

        if k==1:
            c = 0
            m = 0
            y = 0
        else:
            c = (1-r-k)/(1-k)
            m = (1-g-k)/(1-k)
            y = (1-b-k)/(1-k)

        c = round(c*100)
        m = round(m*100)
        y = round(y*100)
        k = round(k*100)

        if issubclass(type(self), colour.ColourAlpha):
            return col_cmyk.cmyka(c, m, y, k, self.a)
        else:
            return col_cmyk.cmyk(c, m, y, k)

    def cmy(self):
        c = 255 - self.r
        m = 255 - self.g
        y = 255 - self.b

        if issubclass(type(self), colour.ColourAlpha):
            return col_cmy.cmya(c, m, y, self.a)
        else:
            return col_cmy.cmy(c, m, y)


class rgba(rgb, colour.ColourAlpha):
    def __init__(self, *v):
        rgb.__init__(self, *v[:-1])
        colour.ColourAlpha.__init__(self, v[-1])

    def __iter__(self):
        t = (self.r, self.g, self.b, self.a)
        for i in t:
            yield i


import importlib
import re
import json
import tp24.internal as internal
import tp24.errors as errors
import tp24.tools as tools

class Colour:
    def __add__(self, other):
        other = internal.samemodel(self, other)
        sv, ov = internal.tuplify(self, other)
        new = tuple(a+b for a, b in zip(sv, ov))
        oc = internal.getclass(self, other)
        new = tuple(c if r>=len(oc.RANGE) or c<=oc.RANGE[r] else oc.RANGE[r] for r, c in enumerate(new)) 
        if issubclass(type(self), ColourAlpha):
            new = list(new)
            new[-1] = 100 if new[-1] > 100 else 0 if new[-1] < 0 else new[-1]
            new = tuple(new)
        return oc(*new)

    def __sub__(self, other):
        other = internal.samemodel(self, other)
        sv, ov = internal.tuplify(self, other)
        new = tuple(a-b for a, b in zip(sv, ov))
        oc = internal.getclass(self, other)
        new = tuple(c if r>=len(oc.RANGE) or c>=0 else 0 for r, c in enumerate(new)) 
        if issubclass(type(self), ColourAlpha):
            new = list(new)
            new[-1] = 100 if new[-1] > 100 else 0 if new[-1] < 0 else new[-1]
            new = tuple(new)
        return oc(*new)

    def __mul__(self, other):
        return tools.gradient(self, other)

    def __repr__(self):
        model = type(self).__name__
        vals = str(tuple(self))
        return model+vals
        
    def hexv(self, compress: bool=False):
        if not type(self).__name__.startswith("rgb"):
            c = self.rgb()
        else:
            c = self

        r = "#"
        for i in tuple(c):
            n = hex(i).split('x')[1]
            if len(n) == 1: n = "0"+n
            r += n

        if compress:
            single = re.search(r"^#(.)\1{5}$", r)
            triple = re.search(r"^#(.)\1(.)\2(.)\3(?:(.)\4)?$", r)
            if single:
                r = "#"+single.group(1)
            elif triple:
                r = "#"
                for i in triple.groups():
                    if i != None: r += i

        return r

    @classmethod
    def from_web(cls, web: str):
        with open("tp24/model/web.json", "r") as f:
            data = json.load(f)
            f.close()
        if not web in data.keys():
            raise ValueError(f"{web} is not a valid web colour")
        return cls.from_hex("#"+data[web])

    @classmethod
    def from_hex(cls, hexc: str):
        c = re.search(r"#(.+)", hexc)
        if not c or not len(c.group(1)) in [1, 3, 4, 6, 8]:
            raise ValueError(f"Hex {hexc} not a valid hex")

        if len(c.group(1)) == 1:
            c = c.group(1)*6
        elif len(c.group(1)) in [3, 4]:
            c = ''.join([char*2 for char in c.group(1)])
        else:
            c = c.group(1)

        channels = tuple(c[i:i+2] for i in range(0, len(c), 2))
        channels = tuple(int('0x'+i, 16) for i in channels)

        import tp24.model.m_rgb as col_rgb

        if issubclass(cls, ColourAlpha):
            channels = list(channels)
            channels[3] *= 100/255
            channels[3] = round(channels[3])
            channels = tuple(channels)
            colour_rgb = col_rgb.rgba(*channels)
        else:
            colour_rgb = col_rgb.rgb(*channels)
        
        if not cls.__name__.startswith("rgb"):
            classname = internal.unalpha(cls.__name__, cls)
            return getattr(colour_rgb, classname)()
        else:
            return colour_rgb

    def inverted(self):
        old = internal.unalpha(tuple(self), self)
        new = tuple(a-b for a, b in zip(self.RANGE, old))
        if issubclass(type(self), ColourAlpha): new = tuple(list(new)+[self.a])

        return type(self)(*new)

    def wheel(self, colours: int, degree: int=None):
        if degree == None:
            degree = 360/(colours+1)
        if degree < 0 or degree > 180:
            raise errors.RangeError(f"Value of degree is {degree} but is not in range of 0 <= d <= 180")
        elif colours <= 0:
            raise errors.RangeError(f"Number of colours is {colours} but is not in range of c > 0")

        if internal.unalpha(type(self).__name__, self) != "hsl":
            c = self.hsl()
        else:
            c = self

        import tp24.model.m_hsl as col_hsl
        cols = []
        for i in range(colours):
            if issubclass(type(c), ColourAlpha):
                cols.append(col_hsl.hsla(*tuple(c)))
            else:
                cols.append(col_hsl.hsl(*tuple(c)))

        for count, i in enumerate(cols):
            direction = count % 2
            if direction == 0: direction = -1
            multiplier = count // 2 + 1
            i.h += direction*degree*multiplier
            while i.h >= 360: i.h -= 360
            while i.h < 0: i.h += 360
        
        if internal.unalpha(type(self).__name__, self) != "hsl":
            #return getattr(t1, internal.unalpha(type(self).__name__, self))(), \
            #       getattr(t2, internal.unalpha(type(self).__name__, self))()
            return tuple(getattr(i, internal.unalpha(type(self).__name__, self))() for i in cols)
        else:
            return tuple(cols)

    def complementary(self):
        return self.wheel(1)[0]

    def triadic(self):
        return self.wheel(2)

    def tetradic(self):
        return self.wheel(3)

    def analogous(self, degree: int=30):
        return self.wheel(2, degree)

    def compound(self, degree: int=30):
        return self.complementary().analogous(degree)

    def add_alpha(self, va: int):
        if issubclass(type(self), ColourAlpha):
            return self
        classname = type(self).__name__+'a'
        modulename = type(self).__name__
        module = importlib.import_module("tp24.model.m_"+modulename)
        vals = tuple(list(self)+[va])
        return getattr(module, classname)(*vals)
    
class ColourAlpha:
    def __init__(self, va: int):
        if va != None and not 0 <= va <= 100:
            raise errors.RangeError(f"Value of A channel is {va} but is not in range of 0 <= a <= 100")
        self.a = va

    def remove_alpha(self):
        classname = type(self).__name__[:-1]
        modulename = type(self).__name__[:-1]
        module = importlib.import_module("tp24.model.m_"+modulename)
        vals = tuple(list(self)[:-1])
        return getattr(module, classname)(*vals)


import tp24.model.colour as colour
import tp24.errors as errors
import tp24.model.m_rgb as col_rgb
import tp24.model.m_hsl as col_hsl

class hsv(colour.Colour):
    RANGE = (360, 100, 100)
    h = None
    s = None
    v = None

    def __init__(self, vh: int, vs: int, vv: int):
        if not 0 <= vh <= self.RANGE[0]:
            raise errors.RangeError(f"Value of H channel is {vh} but is not in range of 0 <= h <= 360")
        elif not 0 <= vs <= self.RANGE[1]:
            raise errors.RangeError(f"Value of S channel is {vs} but is not in range of 0 <= s <= 100")
        elif not 0 <= vv <= self.RANGE[2]:
            raise errors.RangeError(f"Value of V channel is {vv} but is not in range of 0 <= v <= 100")
        self.h = vh
        self.s = vs
        self.v = vv
    
    def __iter__(self):
        t = (self.h, self.s, self.v)
        for i in t:
            yield i

    def rgb(self):
        h = self.h if self.h != self.RANGE[0] else 0
        s = self.s/self.RANGE[1]
        v = self.v/self.RANGE[2]
        c = v*s
        x = c * (1 - abs((h/60)%2 - 1))
        m = v-c

        if 0 <= h < 60:      r, g, b = c, x, 0
        elif 60 <= h < 120:  r, g, b = x, c, 0
        elif 120 <= h < 180: r, g, b = 0, c, x
        elif 180 <= h < 240: r, g, b = 0, x, c
        elif 240 <= h < 300: r, g, b = x, 0, c
        elif 300 <= h < 360: r, g, b = c, 0, x

        r = round((r+m)*255)
        g = round((g+m)*255)
        b = round((b+m)*255)

        if issubclass(type(self), colour.ColourAlpha):
            return col_rgb.rgba(r, g, b, self.a)
        else:
            return col_rgb.rgb(r, g, b)

    def hsl(self):
        h = self.h
        s = self.s/self.RANGE[1]
        v = self.v/self.RANGE[2]

        l = v - (v*s)/2
        if l in [0, 1]: s = 0
        else: s = 2-(2*l)/v

        s = round(s*self.RANGE[1])
        l = round(l*self.RANGE[2])

        if issubclass(type(self), colour.ColourAlpha):
            return col_hsl.hsla(h, s, l, self.a)
        else:
            return col_hsl.hsl(h, s, l)

    def cmyk(self):
        return self.rgb().cmyk()

    def cmy(self):
        return self.rgb().cmy()

class hsva(hsv, colour.ColourAlpha):
    def __init__(self, *v):
        hsv.__init__(self, *v[:-1])
        colour.ColourAlpha.__init__(self, v[-1])

    def __iter__(self):
        t = (self.h, self.s, self.v, self.a)
        for i in t:
            yield i


import tp24.model.colour as colour
import tp24.model.m_rgb as col_rgb
import tp24.model.m_cmy as col_cmy
import tp24.errors as errors

class cmyk(colour.Colour):
    RANGE = (100, 100, 100, 100)
    c = None
    m = None
    y = None
    k = None

    def __init__(self, vc: int, vm: int, vy: int, vk: int):
        if not 0 <= vc <= self.RANGE[0]:
            raise errors.RangeError(f"Value of C channel is {vc} but is not in range of 0 <= C <= 100")
        elif not 0 <= vm <= self.RANGE[1]:
            raise errors.RangeError(f"Value of M channel is {vm} but is not in range of 0 <= m <= 100")
        elif not 0 <= vy <= self.RANGE[2]:
            raise errors.RangeError(f"Value of Y channel is {vy} but is not in range of 0 <= y <= 100")
        elif not 0 <= vk <= self.RANGE[3]:
            raise errors.RangeError(f"Value of K channel is {vk} but is not in range of 0 <= k <= 100")
        self.c = vc
        self.m = vm
        self.y = vy
        self.k = vk
    
    def __iter__(self):
        t = (self.c, self.m, self.y, self.k)
        for i in t:
            yield i

    def rgb(self):
        c = self.c/self.RANGE[0]
        m = self.m/self.RANGE[1]
        y = self.y/self.RANGE[2]
        k = self.k/self.RANGE[3]

        r = 255 * (1-c) * (1-k)
        g = 255 * (1-m) * (1-k)
        b = 255 * (1-y) * (1-k)

        r = round(r)
        g = round(g)
        b = round(b)

        if issubclass(type(self), colour.ColourAlpha):
            return col_rgb.rgba(r, g, b, self.a)
        else:
            return col_rgb.rgb(r, g, b)

    def hsl(self):
        return self.rgb().hsl()

    def hsv(self):
        return self.rgb().hsv()

    def cmy(self):
        if self.k == 0:
            c = self.c
            m = self.m
            y = self.y
        else:
            c = self.c/self.k
            m = self.m/self.k
            y = self.y+self.k

        c *= 100/255
        m *= 100/255
        y *= 100/255

        if issubclass(type(self), colour.ColourAlpha):
            return col_cmy.cmya(c, m, y, self.a)
        else:
            return col_cmy.cmy(c, m, y)

class cmyka(cmyk, colour.ColourAlpha):
    def __init__(self, *v):
        cmyk.__init__(self, *v[:-1])
        colour.ColourAlpha.__init__(self, v[-1])

    def __iter__(self):
        t = (self.c, self.m, self.y, self.k, self.a)
        for i in t:
            yield i


import importlib
import tp24.model.colour as colour

def tuplify(s, o):
    if type(s).__name__ != type(o).__name__:
        pass
    sv = tuple(s)
    ov = tuple(o)
    if len(sv) < len(ov): sv = tuple(list(sv)+[255])
    if len(ov) < len(sv): ov = tuple(list(ov)+[255])
    return sv, ov

def getclass(cls, cls2):
    classname = type(cls).__name__
    if (issubclass(type(cls), colour.ColourAlpha) or issubclass(type(cls2), colour.ColourAlpha)) and not classname.endswith("a"):
        classname += 'a'
    modulename = type(cls).__name__ if not issubclass(type(cls), colour.ColourAlpha) else type(cls).__name__[:-1]
    module = importlib.import_module("tp24.model.m_"+modulename)
    return getattr(module, classname)

def unalpha(i, cls):
    return i[:-1] if issubclass(type(cls), colour.ColourAlpha) else i

def samemodel(a, b):
    if unalpha(type(a).__name__, a) != unalpha(type(b).__name__, b):
        funcname = unalpha(type(a).__name__, a)
        return getattr(b, funcname)()
    else:
        return b


from setuptools import setup
import tp24

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
  name = 'tp24',
  packages = ['tp24'],
  version = tp24.__version__+"",
  license ='gpl-3.0',
  description = 'Python colour library',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = '7d',
  author_email = 'i.third.7d@protonmail.com',
  url = 'https://github.com/iiiii7d/tp24',
  download_url = f'https://github.com/iiiii7d/tp24/archive/refs/tags/v{tp24.__version__}.tar.gz',
  keywords = ['tp24', 'colour', 'color'],
  python_requires='>=3.7',
  package_data={
    'tp24': ['model/*'],
  },
  install_requires=[],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    'Topic :: Utilities'
  ],
)

#commands for upload in case i forget
#python setup.py sdist
#python setup.py bdist_wheel
#twine upload dist/*